/* 
 * $Id: recoder.c,v 1.9 1998/04/24 02:23:55 younggap Exp kangasha $
 *
 * Young Gap Kwon  JAN 30 98
 *
 */

/* Header related to jpeglib */
#include <cdjpeg.h>		/* Common decls for cjpeg/djpeg applications */
#include <jversion.h>		/* for version message */
#include <setjmp.h>             /* We need to jumpout when fatal error occur */


#define write_marker(x) putc(0xFF,outfile);putc((x),outfile); 

#include "squid.h"
#include "recoder.h"


/* Needed by debug */
int opt_syslog_enable = 0;
volatile int unbuffered_logs = 1;
static const char *const w_space = " \t\n\r";
const char *const appname = "recoder";
struct timeval current_time;
time_t squid_curtime;
struct SquidConfig Config;

static struct timeval starttime;

time_t
getCurrentTime(void)
{
#if GETTIMEOFDAY_NO_TZP
    gettimeofday(&current_time);
#else
    gettimeofday(&current_time, NULL);
#endif
    return squid_curtime = current_time.tv_sec;
}


#define TRUE 1
#define FALSE 0
#define boolean char

#define _FOPEN(_name,_f_handle,_perm,_ret_num) \
   if ( (_f_handle=fopen(_name,_perm)) == NULL ) { \
	cleanup(temp_name, temp2_name); \
	return _ret_num; \
    }

/* Global */

int u_level = 0;
int converted = 1; /* 1 if image was converted to progressive, 0 if not. */
jmp_buf env;

/* 
   If something going wrong in jpeg library, we interrupt 
   to come here instead of exiting from program. 
*/

METHODDEF(void)
soft_error_return ( j_common_ptr cinfo)
  {
    jpeg_destroy(cinfo); 
    longjmp(env,-1);
  }

/* 
   This function take infut file pointer and output pointer and 
   generate progressive jpeg output. 
*/
int
jpeg_progress (FILE *input_file,FILE *output_file)
{
  struct jpeg_decompress_struct srcinfo;
  struct jpeg_compress_struct dstinfo;
  struct jpeg_error_mgr jsrcerr, jdsterr;
  int num_level;

  jvirt_barray_ptr * coef_arrays;


  /* Initialize the JPEG decompression object with default error handling. */
  srcinfo.err = jpeg_std_error(&jsrcerr);
  /* Make ruturn rather than exit call */
  jsrcerr.error_exit = soft_error_return;
  /* The other left behind */
  
  jpeg_create_decompress(&srcinfo);
  /* Initialize the JPEG compression object with default error handling. */
  dstinfo.err = jpeg_std_error(&jdsterr);
  jdsterr.error_exit = soft_error_return;
  jpeg_create_compress(&dstinfo);


  jsrcerr.trace_level = jdsterr.trace_level;
  srcinfo.mem->max_memory_to_use = dstinfo.mem->max_memory_to_use;

  /* This should never occured. but needed to check stablility*/
  if ( setjmp(env) != 0) {
    fprintf(stderr,"Error occured!\n");
    return -1;
  }
  /* Specify data source for decompression */
  jpeg_stdio_src(&srcinfo, input_file);

  /* Read file header */
  (void) jpeg_read_header(&srcinfo, TRUE);
  
  /* Read source file as DCT coefficients */
  coef_arrays = jpeg_read_coefficients(&srcinfo);

  /* Initialize destination compression parameters from source values */
  jpeg_copy_critical_parameters(&srcinfo, &dstinfo);

  /* YG: If the original have ncomp less than 3 the scan number is 6*/
  if ( dstinfo.num_components < 3) {
      num_level=6;
  } else { 
    num_level=10;
  }
  /* YG : we only use progressive mode */
  jpeg_simple_progression(&dstinfo);
  
  /* Specify data destination for compression */
  jpeg_stdio_dest(&dstinfo, output_file);

  /* Start compressor */
  jpeg_write_coefficients(&dstinfo, coef_arrays);

  /* ought to copy source comments here... */

  /* Finish compression and release memory */
  jpeg_finish_compress(&dstinfo);
  jpeg_destroy_compress(&dstinfo);
  (void) jpeg_finish_decompress(&srcinfo);
  jpeg_destroy_decompress(&srcinfo);
 
  return num_level;			/* suppress no-return-value warnings */
}


static int 
check_mark (int marker)
{
  switch (marker) {
  case M_SOF0:		/* Baseline */
  case M_SOF1:		/* Extended sequential, Huffman */
  case M_SOF2:		/* Progressive, Huffman */
  case M_SOF3:		/* Lossless, Huffman */
  case M_SOF5:		/* Differential sequential, Huffman */
  case M_SOF6:		/* Differential progressive, Huffman */
  case M_SOF7:		/* Differential lossless, Huffman */
  case M_SOF9:		/* Extended sequential, arithmetic */
  case M_SOF10:		/* Progressive, arithmetic */
  case M_SOF11:		/* Lossless, arithmetic */
  case M_SOF13:		/* Differential sequential, arithmetic */
  case M_SOF14:		/* Differential progressive, arithmetic */
  case M_SOF15:		/* Differential lossless, arithmetic */
  case M_EOI:		/* in case it's a tables-only JPEG stream */
  case M_COM:		/* Comment Marker */
    return 1;           /* TRUE */
  default:
    return 0;
    break;
  }
}

/* 
   It take input and insert comment.
*/

int 
wsoft_title( FILE * infile, FILE * outfile , char *comment_arg)
{
  unsigned int comment_length = 0;
  int ch;
  int oldch=0;
  unsigned int bit2_length;

  comment_length = strlen(comment_arg);
  
  while ( (ch=getc(infile)) != EOF)
    {
      if ( oldch == 0xFF && check_mark(ch)) break;
      fputc(ch,outfile);
      oldch=ch;
    }
  /* Insert the new COM marker */
  write_marker(M_COM);
  bit2_length = comment_length + 2;
  putc(((bit2_length >> 8) & 0xFF),outfile);
  putc( (bit2_length & 0xFF),outfile);
  while (comment_length > 0) {
    putc(*comment_arg++,outfile);
    comment_length--;
  }
  if ( ch == M_COM ){
    comment_length = ( getc(infile) << 8);
    comment_length += (getc(infile));
    comment_length -=2;
    while ( comment_length >0) { getc(infile); comment_length -- ;}
  } else {   write_marker(ch); }
  
  /* Duplicate the remainder of the source file.
   * Note that any COM markers occuring after SOF will not be touched.
   */
  while ( (ch=getc(infile)) != EOF) 
    {
      if ( oldch == 0xFF && ch == M_COM )
	{
	  comment_length = ( getc(infile) << 8);
	  comment_length += (getc(infile));
	  comment_length -=2;
	  while ( comment_length >0) { getc(infile); comment_length -- ;}
	  getc(infile);
	}
      fputc(ch,outfile);
      oldch=ch;
    }
  return 0;			/* suppress no-return-value warnings */
}

/* 
   Small utility function to clean up each temporary file used in jconv.
*/
void 
cleanup(char *temp, char *temp2) 
{
  unlink(temp);
  unlink(temp2);
}

/*
  This function take input as object file of squid cache and
  generate desired output.
  It take file name and #of level(slice) to be removed
*/

int 
jconv(char *f_name, int level)
{
    FILE *fin;
    FILE *fout;
    int i,ch;
    int oldch = 0;
    int SOScount = 0;
    int header_end_line;
    char temp_name[80], temp2_name[80];
    char header[20][80];
    int COMcount = 20;
    char comment[20] = "";
    char wcomment[80];
    boolean SOFTflag = FALSE;
    struct stat finfo;

    /* We use one temporary file to communicate with jpegtrans */
    sprintf(temp_name,"/tmp/%ld.jpg", getpid());
    sprintf(temp2_name,"%s2", temp_name);

    /* collect header information */ 
    _FOPEN(f_name,fin,"r",-13);
    _FOPEN(temp_name,fout,"w",-14);
  
    for(i=0;i<20;i++) {
	fgets(header[i], 80, fin);
	if ( strlen(header[i]) < 3 ) 
	    break;
    }
    header_end_line = i;
    sprintf(comment, " ");
    while ((ch = getc(fin)) != EOF) {
	/* count SOS mark while read input */
	if ( oldch == 0xFF && ch == mcSOS ) 
	    SOScount++;
	if (SOFTflag == FALSE) {
	    if ( oldch == 0xFF && ch == mcCOM) { 
		if ( strstr(comment,"Soft") != NULL )  
		    SOFTflag = TRUE;
		COMcount = 0; 
		sprintf(comment," ");
	    } else {
		if (COMcount < 10) {
		    COMcount++; 
		    sprintf(comment,"%s%c",comment,ch);
		}
	    }
	}
	fputc(ch,fout);
	oldch=ch;
    }
    fclose(fin);
    fclose(fout);
    /* 
       Eventhough we check comment while we divide header,
       We still have remained comment field to check out.
     */
    if ( SOFTflag || (strstr(comment,"Soft") != NULL) ) 
      SOFTflag=TRUE; /* Object is processed */

    /* Progressive transform performed in here */
    if (SOFTflag)  converted=0;
      /*
	Decision maid. The object is already processed.
	Put the output if you want to get some information 
      */
    else if ( SOScount == 10 ) {
      /* The object was already progressive with 10 level.
	 So we don't need to convert this as progressive.
      */
      converted=0;
    } else {
	/* make image file progressive */
      _FOPEN(temp_name,fin,"r",-15);
      _FOPEN(temp2_name,fout,"w",-16);
      if ( (SOScount=jpeg_progress(fin,fout)) < 0) {
	cleanup(temp_name, temp2_name);
	return -3; 
	/* In case of error in jpeg lib. Usually wrong image type*/
      }
      fclose(fin); 
      fclose(fout);
      if (rename(temp2_name,temp_name) != 0 ) {
	cleanup(temp_name, temp2_name);
	return -21;
      }
      converted=1;
      /* SOScount = 10; This value set up by jpeg_progress
	 and if grey scale image this value is 6 */
      /* progressive end */
    }
    /* 
       Now we have progressive object whether we process or original
       object was progressive. temp2_name -> progressive jpeg file
       Now set up u_level here. We received level estimation
       from softd and level should be match with u_level if things going 
       on correctly
    */
    if ( level ) { /* If level is not 0 do degrading */
      SOScount -= (level-1); /* SOScount should have desired level+1 */
      u_level = SOScount - 1; 
      /* This is new level and will cause take off one scan layer. */
      
      /* degrading process */
      if ( SOScount <= 1 ) {
	/* In case we come to lowest level */
	cleanup(temp_name, temp2_name);
	return -2;
      } else  {
	_FOPEN(temp_name,fin,"r",-11);
	_FOPEN(temp2_name,fout,"w",-12);
	i=0;
	while ( (ch=getc(fin)) != EOF) {
	    /* count SOS mark while read input */
	    if ( oldch == 0xFF && ch == mcSOS ) i++;
	    if ( i >= SOScount ) 
	      { 
		putc(mcEOI,fout); 
		fclose(fout); 
		break;
	      }
	    fputc(ch,fout);
	    oldch=ch;
	  }
	} /* degrading end here */
	/* Write comment */
	fclose(fin);
	fclose(fout);
	if (rename(temp2_name,temp_name) != 0 ) {
	  cleanup(temp_name, temp2_name);
	  return -22;
	}
      }
    else
      {
	u_level= SOScount;
      }
  
    _FOPEN(f_name,fout,"w",-17);
    stat(temp_name,&finfo); /* get st_size from striped file*/
    for(i=0;i<=header_end_line;i++) {
	if ( strstr(header[i],"Content-Length:") != NULL )
	    sprintf(header[i],"Content-Length: %ld\x0d\x0a", finfo.st_size);
	fputs(header[i], fout);
    }
    _FOPEN(temp_name,fin,"r",-18);
    sprintf(wcomment,"Soft : Processed by Soft cache Quality=%d%%",u_level*10);
    wsoft_title(fin,fout,wcomment);
    fclose(fin); 
    fclose(fout);
    stat(f_name, &finfo);
    /* clean up remaining temporary file */
    cleanup(temp_name, temp2_name);
    return finfo.st_size;
    /* Everything was successful so return new object size.
       Caution: This is object size not image size. */
}

/* 
   Main
   take argument from standard input formated as [filaneme level].
   output has form [object_size u_level converted]
*/
int 
main(int argc, char **argv)
{
    int object_size, levels;
    static char inbuf[2048];
    char *t;
    static char filename[2048];
    struct timeval start_time;
    
    getCurrentTime();
    starttime = current_time;
    _db_init(NULL, "ALL,1");

    for (;;) {
	if (fgets(inbuf, 2048, stdin) == NULL) {
	    exit(-2);
	}
	
	t = strrchr(inbuf, '\n');
	if (t == NULL)		/* Ignore if no newline */
	    continue;
	*t = '\0';		/* strip NL */
	if ((t = strrchr(inbuf, '\r')) != NULL)
	    *t = '\0';		/* strip CR */
	
	if (strcmp(inbuf, "$shutdown") == 0)
	    exit(0);
	if (strcmp(inbuf, "$hello") == 0) {
	    printf("$alive\n");
	    fflush(stdout);
	    continue;
	}
	
	memset(filename, '\0', 2048);
	sscanf(inbuf, "%s %d", filename, &levels);
	
	getCurrentTime();
	start_time = current_time;
	debug(44, 7, "Recoder looking at image %s\n", filename);
	if ((object_size = jconv(filename, levels)) < 0 ) {
	    if (object_size == -2) {
		debug(44, 2, "%s come to lowest quality and do nothing\n",
		      filename);
		fprintf(stdout,"0 1 0 0\n");
		fflush(stdout);
	    } else if (object_size == -3) {
		debug(44, 2, "%s wrong image type\n", filename);
		fprintf(stdout,"0 -1 -1 0\n");
		fflush(stdout);
	    } else {
		debug(44, 1, "Fatal error. jconv returned %d\n",object_size);
		fprintf(stdout,"0 -1 -1 0\n");
		fflush(stdout);
	    }
	} else {
	    getCurrentTime();
	    fprintf(stdout,"%d %d %d %d\n", object_size, u_level, converted,
		    (int)((current_time.tv_sec - start_time.tv_sec) * 1000 + 
		     (current_time.tv_usec - start_time.tv_usec) / 1000));
	    fflush(stdout);
	}
	debug(44, 6, "Recoder for %s, new size %d\n", filename, object_size);
    }
    return 0;
}

